"""End to End System test"""

import argparse
import json
import sys
from typing import Any, List, Optional

import click
from fido2.hid import CtapHidDevice
from requests import Response

from pyicloud import PyiCloudService
from pyicloud.exceptions import PyiCloudServiceUnavailable
from pyicloud.services.calendar import CalendarObject, CalendarService

END_LIST = "End List\n"
MAX_DISPLAY = 10


def parse_args() -> argparse.Namespace:
    """Parse command line arguments"""
    parser = argparse.ArgumentParser(description="End to End Test of Services")

    parser.add_argument(
        "--username",
        action="store",
        dest="username",
        default="",
        help="Apple ID to Use",
    )
    parser.add_argument(
        "--password",
        action="store",
        dest="password",
        default="",
        help=(
            "Apple ID Password to Use; if unspecified, password will be "
            "fetched from the system keyring."
        ),
    )
    parser.add_argument(
        "--china-mainland",
        action="store_true",
        dest="china_mainland",
        default=False,
        help="If the country/region setting of the Apple ID is China mainland",
    )

    return parser.parse_args()


def handle_2fa(api: PyiCloudService) -> None:
    """Handle two-factor authentication"""
    security_key_names: Optional[List[str]] = api.security_key_names

    if security_key_names:
        print(
            f"Security key confirmation is required. "
            f"Please plug in one of the following keys: {', '.join(security_key_names)}"
        )

        fido2_devices: List[CtapHidDevice] = api.fido2_devices

        print("Available FIDO2 devices:")

        for idx, dev in enumerate(fido2_devices, start=1):
            print(f"{idx}: {dev}")

        choice = click.prompt(
            "Select a FIDO2 device by number",
            type=click.IntRange(1, len(fido2_devices)),
            default=1,
        )
        selected_device: CtapHidDevice = fido2_devices[choice - 1]

        print("Please confirm the action using the security key")

        api.confirm_security_key(selected_device)

    else:
        print("Two-factor authentication required.")
        code: str = input(
            "Enter the code you received of one of your approved devices: "
        )
        result: bool = api.validate_2fa_code(code)
        print(f"Code validation result: {result}")

        if not result:
            print("Failed to verify security code")
            sys.exit(1)

    if not api.is_trusted_session:
        print("Session is not trusted. Requesting trust...")
        result = api.trust_session()
        print(f"Session trust result: {result}")

        if not result:
            print(
                "Failed to request trust. You will likely be prompted for confirmation again in the coming weeks"
            )


def handle_2sa(api: PyiCloudService) -> None:
    """Handle two-step authentication"""
    print("Two-step authentication required. Your trusted devices are:")

    trusted_devices: List[dict[str, Any]] = api.trusted_devices
    for i, device in enumerate(trusted_devices):
        print(
            "  %s: %s"
            % (i, device.get("deviceName", "SMS to %s" % device.get("phoneNumber")))
        )

    device_index: int = click.prompt("Which device would you like to use?", default=0)
    device: dict[str, Any] = trusted_devices[device_index]
    if not api.send_verification_code(device):
        print("Failed to send verification code")
        sys.exit(1)

    code = click.prompt("Please enter validation code")
    if not api.validate_verification_code(device, code):
        print("Failed to verify verification code")
        sys.exit(1)


def get_api() -> PyiCloudService:
    """Get the PyiCloud API"""
    args: argparse.Namespace = parse_args()

    api = PyiCloudService(
        apple_id=args.username,
        password=args.password,
        china_mainland=args.china_mainland,
    )

    if api.requires_2fa:
        handle_2fa(api)

    elif api.requires_2sa:
        handle_2sa(api)

    return api


def display_devices(api: PyiCloudService) -> None:
    """Display device info"""
    print(f"List of devices ({len(api.devices)}):")
    for idx, device in enumerate(api.devices):
        print(f"\t{idx}: {device}")
        if idx >= MAX_DISPLAY - 1:
            break
    print(END_LIST)

    print("First device:")
    print(f"\t Name: {api.iphone}")
    print(f"\t Location: {json.dumps(api.iphone.location, indent=4)}\n")


def display_calendars(api: PyiCloudService) -> None:
    """Display calendar info"""
    calendar_service: CalendarService = api.calendar
    calendars: list[CalendarObject] = calendar_service.get_calendars(as_objs=True)
    print(f"List of calendars ({len(calendars)}):")
    for idx, calendar in enumerate(calendars):
        print(f"\t{idx}: {calendar.title}")
        if idx >= MAX_DISPLAY - 1:
            break
    print(END_LIST)


def display_contacts(api: PyiCloudService) -> None:
    """Display contacts info"""
    contacts = api.contacts.all
    if contacts:
        print(f"List of contacts ({len(contacts)}):")
        for idx, contact in enumerate(contacts):
            print(
                f"\t{idx}: {contact.get('firstName') or contact.get('lastName') or contact.get('companyName')}"
            )
            if idx >= MAX_DISPLAY - 1:
                break
        print(END_LIST)
    else:
        print("No contacts found\n")


def display_drive(api: PyiCloudService) -> None:
    """Display drive info"""
    drive_files: list[str] = api.drive.dir()
    print(f"List of files in iCloud Drive root ({len(drive_files)}):")
    for idx, filename in enumerate(drive_files):
        print(f"\t{idx}: {filename} ({api.drive[filename].type})")
        if idx >= MAX_DISPLAY - 1:
            break
    print(END_LIST)


def display_files(api: PyiCloudService) -> None:
    """Display files info"""
    try:
        files: list[str] = api.files.dir()
        print(f"List of files in iCloud files root ({len(files)}):")
        for idx, filename in enumerate(files):
            print(f"\t{idx}: {filename} ({api.files[filename].type})")
            if idx >= MAX_DISPLAY - 1:
                break
        print(END_LIST)
    except PyiCloudServiceUnavailable as error:
        print(f"Files service not available: {error}\n")


def display_photos(api: PyiCloudService) -> None:
    """Display photo info"""
    print(f"List of photo albums ({len(api.photos.albums)}):")
    for idx, album in enumerate(api.photos.albums):
        print(f"\t{idx}: {album}")
        if idx >= MAX_DISPLAY - 1:
            break
    print(END_LIST)

    print(f"List of ALL PHOTOS ({len(api.photos.all)}):")
    for idx, photo in enumerate(api.photos.all):
        print(f"\t{idx}: {photo.filename} ({photo.item_type})")
        if idx >= MAX_DISPLAY - 1:
            break
    print(END_LIST)


def display_videos(api: PyiCloudService) -> None:
    """Display video info"""

    print(f"List of Videos ({len(api.photos.albums['Videos'])}):")
    for idx, photo in enumerate(api.photos.albums["Videos"]):
        print(f"\t{idx}: {photo.filename} ({photo.item_type})")
        if idx >= MAX_DISPLAY - 1:
            break
    print(END_LIST)


def display_shared_photos(api: PyiCloudService) -> None:
    """Display shared photo info"""

    album = None
    print(f"List of Shared Albums ({len(api.photos.shared_streams)}):")
    for idx, album in enumerate(api.photos.shared_streams):
        print(f"\t{idx}: {album}")
        if idx >= MAX_DISPLAY - 1:
            break
    print(END_LIST)

    if album and api.photos.shared_streams:
        print(
            f"List of Shared Photos [{album}] ({len(api.photos.shared_streams[album])}):"
        )
        for idx, photo in enumerate(api.photos.shared_streams[album]):
            print(f"\t{idx}: {photo.filename} ({photo.item_type})")

            if idx >= MAX_DISPLAY - 1:
                break
        print(END_LIST)


def display_account(api: PyiCloudService) -> None:
    """Display account info"""
    print(f"Account name: {api.account_name}")
    print(f"Account plan: {json.dumps(api.account.summary_plan, indent=4)}")
    print(f"List of Family Member ({len(api.account.family)}):")
    for idx, member in enumerate(api.account.family):
        print(f"\t{idx}: {member}")
        photo: Response = member.get_photo()
        print(f"\t\tPhoto: {photo}")
        print(f"\t\tPhoto type: {photo.headers['Content-Type']}")
        print(f"\t\tPhoto size: {photo.headers['Content-Length']}")
        if idx >= MAX_DISPLAY - 1:
            break
    print(END_LIST)


def display_hidemyemail(api: PyiCloudService) -> None:
    """Display Hide My Email info"""
    print(f"List of Hide My Email ({len(api.hidemyemail)}):")
    for idx, email in enumerate(api.hidemyemail):
        print(
            f"\t{idx}: {email['hme']} ({email['domain']}) Active = {email['isActive']}"
        )
        if idx >= MAX_DISPLAY - 1:
            break
    print(END_LIST)


def main() -> None:
    """main function"""
    api: PyiCloudService = get_api()
    display_hidemyemail(api)
    display_account(api)
    try:
        display_calendars(api)
    except PyiCloudServiceUnavailable as error:
        print(f"Calendar service not available: {error}\n")
    display_files(api)
    display_devices(api)
    display_contacts(api)
    display_drive(api)
    display_photos(api)
    display_videos(api)
    display_shared_photos(api)


if __name__ == "__main__":
    main()



"""Pytest configuration file for the pyicloud package."""

import os
import secrets
from unittest.mock import MagicMock, mock_open, patch

import pytest
from requests.cookies import RequestsCookieJar

from pyicloud.base import PyiCloudService
from pyicloud.services.contacts import ContactsService
from pyicloud.services.drive import COOKIE_APPLE_WEBAUTH_VALIDATE
from pyicloud.services.hidemyemail import HideMyEmailService
from pyicloud.session import PyiCloudSession
from tests import PyiCloudSessionMock
from tests.const_login import LOGIN_WORKING

BUILTINS_OPEN: str = "builtins.open"
EXAMPLE_DOMAIN: str = "https://example.com"


class FileSystemAccessError(Exception):
    """Raised when a test tries to access the file system."""


@pytest.fixture(autouse=True, scope="function")
def mock_mkdir():
    """Mock the mkdir function to prevent file system access."""
    mkdir = os.mkdir

    def my_mkdir(path, *args, **kwargs):
        if "python-test-results" not in path:
            raise FileSystemAccessError(
                f"You should not be creating directories in tests. {path}"
            )
        return mkdir(path, *args, **kwargs)

    with patch("os.mkdir", my_mkdir) as mkdir_mock:
        yield mkdir_mock


@pytest.fixture(autouse=True, scope="session")
def mock_open_fixture():
    """Mock the open function to prevent file system access."""
    builtins_open = open

    def my_open(path, *args, **kwargs):
        if "python-test-results" not in path:
            raise FileSystemAccessError(
                f"You should not be opening files in tests. {path}"
            )
        return builtins_open(path, *args, **kwargs)

    with patch(BUILTINS_OPEN, my_open) as open_mock:
        yield open_mock


@pytest.fixture
def pyicloud_service() -> PyiCloudService:
    """Create a PyiCloudService instance with mocked authenticate method."""
    with (
        patch("pyicloud.base.PyiCloudService.authenticate") as mock_authenticate,
        patch(BUILTINS_OPEN, new_callable=mock_open),
    ):
        # Mock the authenticate method during initialization
        mock_authenticate.return_value = None
        service = PyiCloudService("test@example.com", secrets.token_hex(32))
        return service


@pytest.fixture
def pyicloud_service_working(pyicloud_service: PyiCloudService) -> PyiCloudService:  # pylint: disable=redefined-outer-name
    """Set the service to a working state."""
    pyicloud_service.data = LOGIN_WORKING
    pyicloud_service._webservices = LOGIN_WORKING["webservices"]  # pylint: disable=protected-access
    with patch(BUILTINS_OPEN, new_callable=mock_open):
        pyicloud_service.session = PyiCloudSessionMock(
            pyicloud_service,
            "",
            cookie_directory="",
        )
        pyicloud_service.session._data = {"session_token": "valid_token"}  # pylint: disable=protected-access
    return pyicloud_service


@pytest.fixture
def pyicloud_session(pyicloud_service_working: PyiCloudService) -> PyiCloudSession:  # pylint: disable=redefined-outer-name
    """Mock the PyiCloudSession class."""
    pyicloud_service_working.session.cookies = MagicMock()
    return pyicloud_service_working.session


@pytest.fixture
def mock_session() -> MagicMock:
    """Fixture to create a mock PyiCloudSession."""
    return MagicMock(spec=PyiCloudSession)


@pytest.fixture
def contacts_service(mock_session: MagicMock) -> ContactsService:  # pylint: disable=redefined-outer-name
    """Fixture to create a ContactsService instance."""
    return ContactsService(
        service_root=EXAMPLE_DOMAIN,
        session=mock_session,
        params={"test_param": "value"},
    )


@pytest.fixture
def mock_photos_service() -> MagicMock:
    """Fixture for mocking PhotosService."""
    service = MagicMock()
    service.service_endpoint = EXAMPLE_DOMAIN
    service.params = {"dsid": "12345"}
    service.session = MagicMock()
    return service


@pytest.fixture
def mock_photo_library(mock_photos_service: MagicMock) -> MagicMock:  # pylint: disable=redefined-outer-name
    """Fixture for mocking PhotoLibrary."""
    library = MagicMock()
    library.service = mock_photos_service
    return library


@pytest.fixture
def hidemyemail_service(mock_session: MagicMock) -> HideMyEmailService:  # pylint: disable=redefined-outer-name
    """Fixture for initializing HideMyEmailService."""
    return HideMyEmailService(EXAMPLE_DOMAIN, mock_session, {"dsid": "12345"})


@pytest.fixture
def mock_service_with_cookies(
    pyicloud_service_working: PyiCloudService,  # pylint: disable=redefined-outer-name
) -> PyiCloudService:
    """Fixture to create a mock PyiCloudService with cookies."""
    jar = RequestsCookieJar()
    jar.set(COOKIE_APPLE_WEBAUTH_VALIDATE, "t=768y9u", domain="icloud.com", path="/")

    # Attach a real CookieJar so code that calls `.cookies.get()` keeps working.
    pyicloud_service_working.session.cookies = jar

    return pyicloud_service_working



"""Calendar service."""

import json
from calendar import monthrange
from dataclasses import asdict, dataclass, field
from datetime import datetime, timedelta
from random import randint
from typing import Any, List, Literal, Optional, TypeVar, Union, cast, overload
from uuid import uuid4

from requests import Response
from tzlocal import get_localzone_name

from pyicloud.services.base import BaseService
from pyicloud.session import PyiCloudSession

T = TypeVar("T")


@dataclass
class EventObject:
    """
    An EventObject represents an event in the Apple Calendar.
    """

    pguid: str
    title: str = "New Event"
    start_date: datetime = datetime.today()
    end_date: datetime = datetime.today() + timedelta(minutes=60)
    local_start_date: Optional[datetime] = None
    local_end_date: Optional[datetime] = None
    duration: int = field(init=False)
    icon: int = 0
    change_recurring: Optional[str] = None
    tz: str = "US/Pacific"
    guid: str = ""  # event identifier
    location: str = ""
    extended_details_are_included: bool = True
    recurrence_exception: bool = False
    recurrence_master: bool = False
    has_attachments: bool = False
    all_day: bool = False
    is_junk: bool = False
    etag: Optional[str] = None

    invitees: List[str] = field(init=False, default_factory=list)

    def __post_init__(self) -> None:
        if not self.local_start_date:
            self.local_start_date = self.start_date
        if not self.local_end_date:
            self.local_end_date = self.end_date

        if not self.guid:
            self.guid = str(uuid4()).upper()

        self.duration = int(
            (self.end_date.timestamp() - self.start_date.timestamp()) / 60
        )

    @property
    def request_data(self) -> dict[str, Any]:
        """
        Returns the event data in the format required by Apple's calendar.
        """
        event_dict: dict[str, Any] = asdict(self)
        event_dict["startDate"] = self.dt_to_list(self.start_date)
        event_dict["endDate"] = self.dt_to_list(self.end_date, False)
        if self.local_start_date:
            event_dict["localStartDate"] = self.dt_to_list(self.local_start_date)
        if self.local_end_date:
            event_dict["localEndDate"] = self.dt_to_list(self.local_end_date, False)

        event_dict.pop("start_date", None)
        event_dict.pop("end_date", None)
        event_dict.pop("local_start_date", None)
        event_dict.pop("local_end_date", None)

        data: dict[str, Any] = {
            "Event": event_dict,
            "ClientState": {
                "Collection": [{"guid": self.guid, "ctag": None}],
                "fullState": False,
                "userTime": 1234567890,
                "alarmRange": 60,
            },
        }

        if self.invitees:
            data["Invitee"] = [
                {
                    "guid": email_guid,
                    "pGuid": self.pguid,
                    "role": "REQ-PARTICIPANT",
                    "isOrganizer": False,
                    "email": email_guid.split(":")[-1],
                    "inviteeStatus": "NEEDS-ACTION",
                    "commonName": "",
                    "isMyId": False,
                }
                for email_guid in self.invitees
            ]

        return data

    def dt_to_list(self, dt: datetime, start: bool = True) -> list:
        """
        Converts python datetime object into a list format used
        by Apple's calendar.
        """
        if start:
            minutes: int = dt.hour * 60 + dt.minute
        else:
            minutes = (24 - dt.hour) * 60 + (60 - dt.minute)

        return [
            dt.strftime("%Y%m%d"),
            dt.year,
            dt.month,
            dt.day,
            dt.hour,
            dt.minute,
            minutes,
        ]

    def add_invitees(self, _invitees: Optional[list] = None) -> None:
        """
        Adds a list of emails to invitees in the correct format
        """
        if _invitees:
            self.invitees += ["{}:{}".format(self.guid, email) for email in _invitees]

    def get(self, var: str):
        """Get a variable"""
        return getattr(self, var, None)


@dataclass
class CalendarObject:
    """
    A CalendarObject represents a calendar in the Apple Calendar.
    """

    title: str = "Untitled"
    guid: str = ""
    share_type: Optional[str] = (
        None  # can be (None, 'published', 'shared') where 'published' gens a public caldav link in the response.  Shared is not supported here as it is rather complex.
    )
    symbolic_color: str = "__custom__"
    supported_type: str = "Event"
    object_type: str = "personal"
    share_title: str = ""
    shared_url: str = ""
    color: str = ""
    order: int = 7
    extended_details_are_included: bool = True
    read_only: bool = False
    enabled: bool = True
    ignore_event_updates: Optional[str] = None
    email_notification: Optional[str] = None
    last_modified_date: Optional[str] = None
    me_as_participant: Optional[str] = None
    pre_published_url: Optional[str] = None
    participants: Optional[str] = None
    defer_loading: Optional[str] = None
    published_url: Optional[str] = None
    remove_alarms: Optional[str] = None
    ignore_alarms: Optional[str] = None
    description: Optional[str] = None
    remove_todos: Optional[str] = None
    is_default: Optional[bool] = None
    is_family: Optional[bool] = None
    etag: Optional[str] = None
    ctag: Optional[str] = None

    def __post_init__(self) -> None:
        if not self.guid:
            self.guid = str(uuid4()).upper()

        if not self.color:
            self.color = self.gen_random_color()

    def gen_random_color(self) -> str:
        """
        Creates a random rgbhex color.
        """
        return "#%02x%02x%02x" % tuple([randint(0, 255) for _ in range(3)])

    @property
    def request_data(self) -> dict[str, Any]:
        """Returns the calendar data in the format required by Apple's calendar."""
        data: dict[str, Any] = {
            "Collection": asdict(self),
            "ClientState": {
                "Collection": [],
                "fullState": False,
                "userTime": 1234567890,
                "alarmRange": 60,
            },
        }
        return data


class CalendarService(BaseService):
    """
    The 'Calendar' iCloud service, connects to iCloud and returns events.
    """

    def __init__(
        self, service_root: str, session: PyiCloudSession, params: dict[str, Any]
    ) -> None:
        super().__init__(service_root, session, params)
        self._calendar_endpoint: str = f"{self.service_root}/ca"
        self._calendar_refresh_url: str = f"{self._calendar_endpoint}/events"
        self._calendar_event_detail_url: str = f"{self._calendar_endpoint}/eventdetail"
        self._calendar_collections_url: str = f"{self._calendar_endpoint}/collections"
        self._calendars_url: str = f"{self._calendar_endpoint}/allcollections"

    @property
    def default_params(self) -> dict[str, Any]:
        """Returns the default parameters for the calendar service."""
        today: datetime = datetime.today()
        first_day, last_day = monthrange(today.year, today.month)
        from_dt = datetime(today.year, today.month, first_day)
        to_dt = datetime(today.year, today.month, last_day)
        params = dict(self.params)
        params.update(
            {
                "lang": "en-us",
                "usertz": get_localzone_name(),
                "startDate": from_dt.strftime("%Y-%m-%d"),
                "endDate": to_dt.strftime("%Y-%m-%d"),
            }
        )

        return params

    def obj_from_dict(self, obj: T, _dict) -> T:
        """Creates an object from a dictionary"""
        for key, value in _dict.items():
            setattr(obj, key, value)

        return obj

    def get_ctag(self, guid: str) -> str:
        """Returns the ctag for a given calendar guid"""
        ctag: Optional[str] = None
        for cal in self.get_calendars(as_objs=False):
            if isinstance(cal, CalendarObject) and cal.guid == guid:
                ctag = cal.ctag
            elif isinstance(cal, dict) and cal.get("guid") == guid:
                ctag = cal.get("ctag")

            if ctag:
                return ctag
        raise ValueError("ctag not found.")

    def refresh_client(self, from_dt=None, to_dt=None) -> dict[str, Any]:
        """
        Refreshes the CalendarService endpoint, ensuring that the
        event data is up-to-date. If no 'from_dt' or 'to_dt' datetimes
        have been given, the range becomes this month.
        """
        today: datetime = datetime.today()
        first_day, last_day = monthrange(today.year, today.month)
        if not from_dt:
            from_dt = datetime(today.year, today.month, first_day)
        if not to_dt:
            to_dt = datetime(today.year, today.month, last_day)
        params = dict(self.params)
        params.update(
            {
                "lang": "en-us",
                "usertz": get_localzone_name(),
                "startDate": from_dt.strftime("%Y-%m-%d"),
                "endDate": to_dt.strftime("%Y-%m-%d"),
                "dsid": self.session.service.data["dsInfo"]["dsid"],
            }
        )
        req: Response = self.session.get(self._calendar_refresh_url, params=params)
        return req.json()

    @overload
    def get_calendars(self) -> list[dict[str, Any]]: ...

    @overload
    def get_calendars(self, as_objs: Literal[False]) -> list[dict[str, Any]]: ...

    @overload
    def get_calendars(self, as_objs: Literal[True]) -> list[CalendarObject]: ...

    def get_calendars(
        self, as_objs: Union[Literal[True], Literal[False]] = False
    ) -> Union[list[dict[str, Any]], list[CalendarObject]]:
        """
        Retrieves calendars of this month.
        """
        params: dict[str, Any] = self.default_params
        req: Response = self.session.get(self._calendars_url, params=params)
        response = req.json()
        calendars: list[dict[str, Any]] = response["Collection"]

        if not as_objs and calendars:
            return calendars

        return [self.obj_from_dict(CalendarObject(), cal) for cal in calendars]

    def add_calendar(self, calendar: CalendarObject) -> dict[str, Any]:
        """
        Adds a Calendar to the apple calendar.
        """
        data: dict[str, Any] = calendar.request_data
        params: dict[str, Any] = self.default_params

        req: Response = self.session.post(
            self._calendar_collections_url + f"/{calendar.guid}",
            params=params,
            data=json.dumps(data),
        )
        return req.json()

    def remove_calendar(self, cal_guid: str) -> dict[str, Any]:
        """
        Removes a Calendar from the apple calendar.
        """
        params: dict[str, Any] = self.default_params
        params["methodOverride"] = "DELETE"

        req: Response = self.session.post(
            self._calendar_collections_url + f"/{cal_guid}",
            params=params,
            data=json.dumps({}),
        )
        return req.json()

    def get_events(
        self,
        from_dt: Optional[datetime] = None,
        to_dt: Optional[datetime] = None,
        period: str = "month",
        as_objs: bool = False,
    ) -> list:
        """
        Retrieves events for a given date range, by default, this month.
        """
        today: datetime = datetime.today()
        if period != "month" and from_dt:
            today = datetime(from_dt.year, from_dt.month, from_dt.day)

        if period == "day":
            if not from_dt:
                from_dt = datetime(today.year, today.month, today.day)
            to_dt = from_dt + timedelta(days=1)
        elif period == "week":
            if not from_dt:
                from_dt = datetime(today.year, today.month, today.day) - timedelta(
                    days=today.weekday() + 1
                )
            to_dt = from_dt + timedelta(days=6)

        response: dict[str, Any] = self.refresh_client(from_dt, to_dt)
        events: list = response.get("Event", [])

        if as_objs and events:
            for idx, event in enumerate(events):
                events[idx] = self.obj_from_dict(EventObject(""), event)

        return events

    def get_event_detail(self, pguid, guid, as_obj: bool = False) -> EventObject:
        """
        Fetches a single event's details by specifying a pguid
        (a calendar) and a guid (an event's ID).
        """
        params = dict(self.params)
        params.update(
            {
                "lang": "en-us",
                "usertz": get_localzone_name(),
                "dsid": self.session.service.data["dsInfo"]["dsid"],
            }
        )
        url: str = f"{self._calendar_event_detail_url}/{pguid}/{guid}"
        req: Response = self.session.get(url, params=params)
        response = req.json()
        event = response["Event"][0]

        if as_obj and event:
            event: EventObject = cast(
                EventObject,
                self.obj_from_dict(EventObject(pguid=pguid), event),
            )

        return event

    def add_event(self, event: EventObject) -> dict[str, Any]:
        """
        Adds an Event to a calendar.
        """
        data = event.request_data
        data["ClientState"]["Collection"][0]["ctag"] = self.get_ctag(event.guid)
        params = self.default_params

        req: Response = self.session.post(
            self._calendar_refresh_url + f"/{event.pguid}/{event.guid}",
            params=params,
            data=json.dumps(data),
        )
        return req.json()

    def remove_event(self, event: EventObject) -> dict[str, Any]:
        """
        Removes an Event from a calendar. The calendar's guid corresponds to the EventObject's pGuid
        """
        data = event.request_data
        data["ClientState"]["Collection"][0]["ctag"] = self.get_ctag(event.guid)
        data["Event"] = {}

        params: dict[str, Any] = self.default_params
        params["methodOverride"] = "DELETE"
        if not getattr(event, "etag", None):
            event.etag = self.get_event_detail(
                event.pguid, event.guid, as_obj=False
            ).get("etag")
        params["ifMatch"] = event.etag

        req: Response = self.session.post(
            self._calendar_refresh_url + f"/{event.pguid}/{event.guid}",
            params=params,
            data=json.dumps(data),
        )
        return req.json()



"""PhotoLibrary tests."""

import json
from typing import Any
from unittest.mock import MagicMock, mock_open, patch

import pytest

from pyicloud.const import CONTENT_TYPE, CONTENT_TYPE_TEXT
from pyicloud.exceptions import (
    PyiCloudAPIResponseException,
    PyiCloudServiceNotActivatedException,
)
from pyicloud.services.photos import (
    AlbumContainer,
    BasePhotoAlbum,
    BasePhotoLibrary,
    DirectionEnum,
    PhotoAlbum,
    PhotoLibrary,
    PhotosService,
    PhotoStreamLibrary,
    SharedPhotoStreamAlbum,
    SmartAlbumEnum,
)


def test_photo_library_initialization(mock_photos_service: MagicMock) -> None:
    """Tests initialization of PhotoLibrary."""
    mock_photos_service.session.post.side_effect = [
        MagicMock(
            json=MagicMock(
                return_value={
                    "records": [
                        {
                            "fields": {
                                "state": {"value": "FINISHED"},
                            },
                        }
                    ],
                }
            )
        ),
    ]
    library = PhotoLibrary(
        service=mock_photos_service,
        zone_id={"zoneName": "PrimarySync"},
        upload_url="https://upload.example.com",
    )
    assert library.zone_id == {"zoneName": "PrimarySync"}
    assert library.url == ("https://example.com/records/query?dsid=12345")


def test_photo_library_indexing_not_finished(mock_photos_service: MagicMock) -> None:
    """Tests exception when indexing is not finished."""
    mock_photos_service.session.post.return_value.json.return_value = {
        "records": [
            {
                "fields": {
                    "state": {"value": "NOT_FINISHED"},
                },
            }
        ]
    }
    with pytest.raises(PyiCloudServiceNotActivatedException):
        PhotoLibrary(
            service=mock_photos_service,
            zone_id={"zoneName": "PrimarySync"},
            upload_url="https://upload.example.com",
        )


def test_fetch_folders(mock_photos_service: MagicMock) -> None:
    """Tests the _fetch_folders method."""
    mock_photos_service.session.post.side_effect = [
        MagicMock(
            json=MagicMock(
                return_value={
                    "records": [
                        {
                            "fields": {
                                "state": {"value": "FINISHED"},
                            },
                        }
                    ],
                }
            )
        ),
        MagicMock(
            json=MagicMock(
                return_value={
                    "records": [
                        {
                            "recordName": "folder1",
                            "fields": {
                                "albumNameEnc": {"value": "Zm9sZGVyMQ=="},
                                "isDeleted": {"value": False},
                            },
                        }
                    ]
                }
            )
        ),
    ]

    library = PhotoLibrary(
        service=mock_photos_service,
        zone_id={"zoneName": "PrimarySync"},
        upload_url="https://upload.example.com",
    )
    library.SMART_ALBUMS = {}
    albums: list[BasePhotoAlbum] = list(library.albums.values())

    assert len(albums) == 1
    assert albums[0].name == "folder1"


def test_get_albums(mock_photos_service: MagicMock) -> None:
    """Tests the _get_albums method."""
    mock_photos_service.session.post.side_effect = [
        MagicMock(
            json=MagicMock(
                return_value={
                    "records": [
                        {
                            "fields": {
                                "state": {"value": "FINISHED"},
                            },
                        }
                    ],
                }
            )
        ),
        MagicMock(
            json=MagicMock(
                return_value={
                    "records": [
                        {
                            "recordName": "folder1",
                            "fields": {
                                "albumNameEnc": {"value": "Zm9sZGVyMQ=="},
                                "isDeleted": {"value": False},
                            },
                        }
                    ]
                }
            )
        ),
    ]

    library = PhotoLibrary(
        service=mock_photos_service,
        zone_id={"zoneName": "PrimarySync"},
        upload_url="https://upload.example.com",
    )
    albums: AlbumContainer = library.albums
    assert SmartAlbumEnum.ALL_PHOTOS in albums
    assert "folder1" in albums
    assert albums["folder1"].name == "folder1"
    assert albums["folder1"].direction == DirectionEnum.ASCENDING


def test_upload_file_success(mock_photos_service: MagicMock) -> None:
    """Tests the upload_file method for successful upload."""
    mock_photos_service.params = {"dsid": "12345"}
    mock_photos_service.session.post.side_effect = [
        MagicMock(
            json=MagicMock(
                return_value={
                    "records": [
                        {
                            "fields": {
                                "state": {"value": "FINISHED"},
                            },
                        }
                    ],
                }
            )
        ),
        MagicMock(
            json=MagicMock(
                return_value={
                    "records": [
                        {
                            "recordName": "uploaded_photo",
                            "recordType": "CPLAsset",
                        }
                    ]
                }
            )
        ),
    ]

    library = PhotoLibrary(
        service=mock_photos_service,
        zone_id={"zoneName": "PrimarySync"},
        upload_url="https://upload.example.com",
    )

    with patch("builtins.open", mock_open(read_data=b"file_content")):
        record_name: dict[str, Any] = library.upload_file("test_photo.jpg")

    assert record_name == "uploaded_photo"
    mock_photos_service.session.post.assert_called_with(
        url="https://upload.example.com/upload",
        data=b"file_content",
        params={"filename": "test_photo.jpg", "dsid": "12345"},
    )


def test_upload_file_with_errors(mock_photos_service: MagicMock) -> None:
    """Tests the upload_file method when the response contains errors."""
    mock_photos_service.params = {"dsid": "12345"}
    mock_photos_service.session.post.side_effect = [
        MagicMock(
            json=MagicMock(
                return_value={
                    "records": [
                        {
                            "fields": {
                                "state": {"value": "FINISHED"},
                            },
                        }
                    ],
                }
            )
        ),
        MagicMock(
            json=MagicMock(
                return_value={
                    "errors": [
                        {
                            "code": "UPLOAD_ERROR",
                            "message": "Upload failed",
                        },
                    ],
                }
            )
        ),
    ]

    library = PhotoLibrary(
        service=mock_photos_service,
        zone_id={"zoneName": "PrimarySync"},
        upload_url="https://upload.example.com",
    )

    with patch("builtins.open", mock_open(read_data=b"file_content")):
        with pytest.raises(PyiCloudAPIResponseException) as exc_info:
            library.upload_file("test_photo.jpg")

    assert "UPLOAD_ERROR" in str(exc_info.value)
    mock_photos_service.session.post.assert_called_with(
        url="https://upload.example.com/upload",
        data=b"file_content",
        params={"filename": "test_photo.jpg", "dsid": "12345"},
    )


def test_upload_file_no_records(mock_photos_service: MagicMock) -> None:
    """Tests the upload_file method when no records are returned."""
    mock_photos_service.params = {"dsid": "12345"}
    mock_photos_service.session.post.side_effect = [
        MagicMock(
            json=MagicMock(
                return_value={
                    "records": [
                        {
                            "fields": {
                                "state": {"value": "FINISHED"},
                            },
                        }
                    ],
                }
            )
        ),
        MagicMock(
            json=MagicMock(
                return_value={
                    "records": [],
                }
            )
        ),
    ]
    library = PhotoLibrary(
        service=mock_photos_service,
        zone_id={"zoneName": "PrimarySync"},
        upload_url="https://upload.example.com",
    )

    with patch("builtins.open", mock_open(read_data=b"file_content")):
        with pytest.raises(IndexError):
            library.upload_file("test_photo.jpg")

    mock_photos_service.session.post.assert_called_with(
        url="https://upload.example.com/upload",
        data=b"file_content",
        params={"filename": "test_photo.jpg", "dsid": "12345"},
    )


def test_fetch_folders_multiple_pages(mock_photos_service: MagicMock) -> None:
    """Tests _fetch_folders with multiple pages of results."""
    mock_photos_service.session.post.side_effect = [
        MagicMock(
            json=MagicMock(
                return_value={
                    "records": [
                        {
                            "fields": {
                                "state": {"value": "FINISHED"},
                            },
                        },
                    ],
                }
            )
        ),
        MagicMock(
            json=MagicMock(
                return_value={
                    "records": [
                        {
                            "recordName": "folder1",
                            "fields": {
                                "albumNameEnc": {"value": "Zm9sZGVyMQ=="},
                                "isDeleted": {"value": False},
                            },
                        }
                    ],
                    "continuationMarker": "marker1",
                }
            )
        ),
        MagicMock(
            json=MagicMock(
                return_value={
                    "records": [
                        {
                            "recordName": "folder2",
                            "fields": {
                                "albumNameEnc": {"value": "Zm9sZGVyMg=="},
                                "isDeleted": {"value": False},
                            },
                        }
                    ]
                }
            )
        ),
    ]

    library = PhotoLibrary(
        service=mock_photos_service,
        zone_id={"zoneName": "PrimarySync"},
        upload_url="https://upload.example.com",
    )
    library.SMART_ALBUMS = {}
    albums: list[BasePhotoAlbum] = list(library.albums.values())
    assert len(albums) == 2
    assert albums[0].name == "folder1"
    assert albums[1].name == "folder2"
    mock_photos_service.session.post.assert_called()


def test_fetch_folders_skips_deleted_folders(mock_photos_service: MagicMock) -> None:
    """Tests _fetch_folders skips folders marked as deleted."""
    mock_photos_service.session.post.side_effect = [
        MagicMock(
            json=MagicMock(
                return_value={
                    "records": [
                        {
                            "fields": {
                                "state": {"value": "FINISHED"},
                            },
                        },
                    ],
                }
            )
        ),
        MagicMock(
            json=MagicMock(
                return_value={
                    "continuationMarker": "marker1",
                    "records": [
                        {
                            "recordName": "folder1",
                            "fields": {
                                "albumNameEnc": {"value": "Zm9sZGVyMQ=="},
                                "isDeleted": {"value": True},
                            },
                        },
                    ],
                }
            )
        ),
        MagicMock(
            json=MagicMock(
                return_value={
                    "records": [
                        {
                            "recordName": "folder2",
                            "fields": {
                                "albumNameEnc": {"value": "Zm9sZGVyMg=="},
                                "isDeleted": {"value": False},
                            },
                        },
                    ]
                },
            )
        ),
    ]

    library = PhotoLibrary(
        service=mock_photos_service,
        zone_id={"zoneName": "PrimarySync"},
        upload_url="https://upload.example.com",
    )
    library.SMART_ALBUMS = {}
    albums: list[BasePhotoAlbum] = list(library.albums.values())

    assert len(albums) == 1
    assert albums[0].name == "folder2"
    mock_photos_service.session.post.assert_called()


def test_fetch_folders_no_records(mock_photos_service: MagicMock) -> None:
    """Tests _fetch_folders when no records are returned."""
    mock_photos_service.session.post.side_effect = [
        MagicMock(
            json=MagicMock(
                return_value={
                    "records": [
                        {
                            "fields": {
                                "state": {"value": "FINISHED"},
                            },
                        },
                    ],
                }
            )
        ),
        MagicMock(
            json=MagicMock(
                return_value={
                    "records": [],
                }
            )
        ),
    ]

    library = PhotoLibrary(
        service=mock_photos_service,
        zone_id={"zoneName": "PrimarySync"},
        upload_url="https://upload.example.com",
    )
    library.SMART_ALBUMS = {}
    albums: AlbumContainer = library.albums

    assert len(albums) == 0
    mock_photos_service.session.post.assert_called()


def test_fetch_folders_handles_missing_fields(mock_photos_service: MagicMock) -> None:
    """Tests _fetch_folders handles records with missing fields."""
    mock_photos_service.session.post.side_effect = [
        MagicMock(
            json=MagicMock(
                return_value={
                    "records": [
                        {
                            "fields": {
                                "state": {"value": "FINISHED"},
                            },
                        },
                    ],
                }
            )
        ),
        MagicMock(
            json=MagicMock(
                return_value={
                    "records": [
                        {
                            "recordName": "folder1",
                            "fields": {
                                "isDeleted": {"value": False},
                            },
                        }
                    ]
                }
            )
        ),
    ]

    library = PhotoLibrary(
        service=mock_photos_service,
        zone_id={"zoneName": "PrimarySync"},
        upload_url="https://upload.example.com",
    )
    library.SMART_ALBUMS = {}
    albums: AlbumContainer = library.albums

    assert len(albums) == 0
    mock_photos_service.session.post.assert_called()


def test_base_photo_album_initialization(mock_photo_library: MagicMock) -> None:
    """Tests initialization of BasePhotoAlbum."""
    album = BasePhotoAlbum(
        library=mock_photo_library,
        name="Test Album",
        list_type="CPLAssetAndMasterByAssetDate",
        asset_type=MagicMock,
        page_size=50,
        direction=DirectionEnum.ASCENDING,
    )
    assert album.name == "Test Album"
    assert album.service == mock_photo_library.service
    assert album.page_size == 50
    assert album.direction == DirectionEnum.ASCENDING
    assert album.list_type == "CPLAssetAndMasterByAssetDate"
    assert album.asset_type == MagicMock


def test_base_photo_album_parse_response() -> None:
    """Tests the _parse_response method."""
    album = BasePhotoAlbum(
        library=MagicMock(),
        name="Test Album",
        list_type="CPLAssetAndMasterByAssetDate",
        asset_type=MagicMock,
    )
    response = {
        "records": [
            {
                "recordType": "CPLAsset",
                "fields": {"masterRef": {"value": {"recordName": "master1"}}},
            },
            {
                "recordType": "CPLMaster",
                "recordName": "master1",
            },
        ]
    }
    asset_records, master_records = album._parse_response(response)  # pylint: disable=protected-access
    assert "master1" in asset_records
    assert len(master_records) == 1
    assert master_records[0]["recordName"] == "master1"


def test_base_photo_album_get_photos_at(mock_photo_library: MagicMock) -> None:
    """Tests the _get_photos_at method."""
    mock_photo_library.service.session.post.return_value.json.side_effect = [
        {
            "records": [
                {
                    "recordType": "CPLAsset",
                    "fields": {"masterRef": {"value": {"recordName": "master1"}}},
                },
                {
                    "recordType": "CPLMaster",
                    "recordName": "master1",
                },
            ]
        },
        {
            "records": [],
        },
    ]
    album = PhotoAlbum(
        library=mock_photo_library,
        name="Test Album",
        list_type="CPLAssetAndMasterByAssetDate",
        obj_type="MagicMock",
        direction=DirectionEnum.ASCENDING,
        page_size=10,
        url="https://example.com/records/query?dsid=12345",
    )
    photos = list(album.photos)
    assert len(photos) == 1
    mock_photo_library.service.session.post.assert_called()


def test_base_photo_album_len(mock_photos_service: MagicMock) -> None:
    """Tests the __len__ method."""
    album = BasePhotoAlbum(
        library=mock_photos_service,
        name="Test Album",
        list_type="CPLAssetAndMasterByAssetDate",
        asset_type=MagicMock,
    )
    album._get_len = MagicMock(return_value=42)  # pylint: disable=protected-access
    assert len(album) == 42
    album._get_len.assert_called_once()  # pylint: disable=protected-access


def test_base_photo_album_iter(mock_photo_library: MagicMock) -> None:
    """Tests the __iter__ method."""
    mock_photo_library.service.session.post.return_value.json.side_effect = [
        {
            "records": [
                {
                    "recordType": "CPLAsset",
                    "fields": {"masterRef": {"value": {"recordName": "master1"}}},
                },
                {
                    "recordType": "CPLMaster",
                    "recordName": "master1",
                },
            ]
        },
        {
            "records": [],
        },
    ]
    album = PhotoAlbum(
        library=mock_photo_library,
        name="Test Album",
        list_type="CPLAssetAndMasterByAssetDate",
        obj_type="MagicMock",
        direction=DirectionEnum.ASCENDING,
        page_size=10,
        url="https://example.com/records/query?dsid=12345",
    )
    photos = list(iter(album))
    assert len(photos) == 1
    mock_photo_library.service.session.post.assert_called()


def test_base_photo_album_str() -> None:
    """Tests the __str__ method."""
    album = BasePhotoAlbum(
        library=MagicMock(),
        name="Test Album",
        list_type="CPLAssetAndMasterByAssetDate",
        asset_type=MagicMock,
    )
    assert str(album) == "Test Album"


def test_base_photo_album_repr() -> None:
    """Tests the __repr__ method."""
    album = BasePhotoAlbum(
        library=MagicMock(),
        name="Test Album",
        list_type="CPLAssetAndMasterByAssetDate",
        asset_type=MagicMock,
    )
    assert repr(album) == "<BasePhotoAlbum: 'Test Album'>"


def test_photos_service_initialization(mock_photos_service: MagicMock) -> None:
    """Tests initialization of PhotosService."""
    mock_photos_service.session.post.return_value.json.return_value = {
        "records": [
            {
                "fields": {
                    "state": {"value": "FINISHED"},
                },
            }
        ]
    }
    photos_service = PhotosService(
        service_root="https://example.com",
        session=mock_photos_service.session,
        params={"dsid": "12345"},
        upload_url="https://upload.example.com",
        shared_streams_url="https://shared.example.com",
    )
    assert photos_service.service_endpoint == (
        "https://example.com/database/1/com.apple.photos.cloud/production/private"
    )
    assert isinstance(photos_service._root_library, PhotoLibrary)  # pylint: disable=protected-access
    assert isinstance(photos_service._shared_library, PhotoStreamLibrary)  # pylint: disable=protected-access
    assert photos_service.params["remapEnums"] is True
    assert photos_service.params["getCurrentSyncToken"] is True


def test_photos_service_libraries(mock_photos_service: MagicMock) -> None:
    """Tests the libraries property."""
    mock_photos_service.session.post.return_value.json.side_effect = [
        {
            "records": [
                {
                    "fields": {
                        "state": {"value": "FINISHED"},
                    },
                }
            ]
        },
        {
            "zones": [
                {"zoneID": {"zoneName": "CustomZone"}, "deleted": False},
            ]
        },
        {
            "records": [
                {
                    "fields": {
                        "state": {"value": "FINISHED"},
                    },
                }
            ]
        },
    ]
    photos_service = PhotosService(
        service_root="https://example.com",
        session=mock_photos_service.session,
        params={"dsid": "12345"},
        upload_url="https://upload.example.com",
        shared_streams_url="https://shared.example.com",
    )
    libraries = photos_service.libraries
    assert "root" in libraries
    assert "shared" in libraries
    assert "CustomZone" in libraries
    assert isinstance(libraries["root"], PhotoLibrary)
    assert isinstance(libraries["shared"], PhotoStreamLibrary)
    assert isinstance(libraries["CustomZone"], PhotoLibrary)
    mock_photos_service.session.post.assert_called_with(
        url="https://example.com/database/1/com.apple.photos.cloud/production/private/records/query?dsid=12345&remapEnums=True&getCurrentSyncToken=True",
        data=json.dumps(
            {
                "query": {"recordType": "CheckIndexingState"},
                "zoneID": {"zoneName": "CustomZone"},
            }
        ),
        headers={CONTENT_TYPE: CONTENT_TYPE_TEXT},
    )


def test_photos_service_libraries_cached(mock_photos_service: MagicMock) -> None:
    """Tests that libraries are cached after the first access."""
    mock_photos_service.session.post.return_value.json.return_value = {
        "records": [
            {
                "fields": {
                    "state": {"value": "FINISHED"},
                },
            }
        ]
    }
    photos_service = PhotosService(
        service_root="https://example.com",
        session=mock_photos_service.session,
        params={"dsid": "12345"},
        upload_url="https://upload.example.com",
        shared_streams_url="https://shared.example.com",
    )
    mock_libraries = {"cached": MagicMock(spec=PhotoLibrary)}
    photos_service._libraries = mock_libraries  # type: ignore # pylint: disable=protected-access
    libraries: dict[str, BasePhotoLibrary] = photos_service.libraries
    assert libraries == mock_libraries
    mock_photos_service.session.post.assert_called_once()


def test_photos_service_albums(mock_photos_service: MagicMock) -> None:
    """Tests the albums property."""
    mock_photos_service.session.post.return_value.json.return_value = {
        "records": [
            {
                "fields": {
                    "state": {"value": "FINISHED"},
                },
            }
        ]
    }
    photos_service = PhotosService(
        service_root="https://example.com",
        session=mock_photos_service.session,
        params={"dsid": "12345"},
        upload_url="https://upload.example.com",
        shared_streams_url="https://shared.example.com",
    )
    albums: AlbumContainer = photos_service.albums
    assert isinstance(albums, AlbumContainer)
    assert SmartAlbumEnum.ALL_PHOTOS in albums
    mock_photos_service.session.post.assert_called()


def test_photos_service_shared_streams(mock_photos_service: MagicMock) -> None:
    """Tests the shared_streams property."""
    mock_photos_service.session.post.return_value.json.side_effect = [
        {
            "records": [
                {
                    "fields": {
                        "state": {"value": "FINISHED"},
                    },
                }
            ]
        },
        {
            "albums": [
                {
                    "albumlocation": "https://shared.example.com/album/",
                    "albumctag": "ctag",
                    "albumguid": "guid",
                    "ownerdsid": "owner",
                    "attributes": {
                        "name": "Shared Album",
                        "creationDate": "1234567890",
                        "allowcontributions": True,
                        "ispublic": False,
                    },
                    "sharingtype": "owned",
                    "iswebuploadsupported": True,
                }
            ]
        },
    ]
    photos_service = PhotosService(
        service_root="https://example.com",
        session=mock_photos_service.session,
        params={"dsid": "12345"},
        upload_url="https://upload.example.com",
        shared_streams_url="https://shared.example.com",
    )
    shared_streams: AlbumContainer = photos_service.shared_streams
    assert isinstance(shared_streams, AlbumContainer)
    assert "Shared Album" in shared_streams
    assert isinstance(shared_streams["Shared Album"], SharedPhotoStreamAlbum)
    mock_photos_service.session.post.assert_called()



"""Test calendar service"""

from datetime import datetime
from typing import Any
from unittest.mock import MagicMock, patch

from requests import Response

from pyicloud.services.calendar import CalendarObject, CalendarService, EventObject
from pyicloud.session import PyiCloudSession


def test_event_object_initialization() -> None:
    """Test EventObject initialization and default values."""
    event = EventObject(pguid="calendar123")
    assert event.pguid == "calendar123"
    assert event.title == "New Event"
    assert event.duration == 60
    assert event.tz == "US/Pacific"
    assert event.guid != ""


def test_event_object_request_data() -> None:
    """Test EventObject request_data property."""
    event = EventObject(pguid="calendar123")
    data: dict[str, Any] = event.request_data
    assert "Event" in data
    assert "ClientState" in data
    assert data["Event"]["title"] == "New Event"
    assert "pguid" in data["Event"]
    assert data["Event"]["pguid"] == "calendar123"
    assert "guid" in data["Event"]
    assert "Collection" in data["ClientState"]


def test_event_object_dt_to_list() -> None:
    """Test EventObject dt_to_list method."""
    event = EventObject(pguid="calendar123")
    dt = datetime(2023, 1, 1, 12, 30)
    result = event.dt_to_list(dt)
    assert result == ["20230101", 2023, 1, 1, 12, 30, 750]


def test_event_object_add_invitees() -> None:
    """Test EventObject add_invitees method."""
    event = EventObject(pguid="calendar123")
    event.add_invitees(["test@example.com", "user@example.com"])
    assert len(event.invitees) == 2
    assert f"{event.guid}:test@example.com" == event.invitees[0]
    assert f"{event.guid}:user@example.com" == event.invitees[1]


def test_calendar_object_initialization() -> None:
    """Test CalendarObject initialization and default values."""
    calendar = CalendarObject(title="My Calendar")
    assert calendar.title == "My Calendar"
    assert calendar.guid != ""
    assert calendar.color.startswith("#")


def test_calendar_object_request_data() -> None:
    """Test CalendarObject request_data property."""
    calendar = CalendarObject(title="My Calendar")
    data: dict[str, Any] = calendar.request_data
    assert "Collection" in data
    assert data["Collection"]["title"] == "My Calendar"
    assert "ClientState" in data
    assert "guid" in data["Collection"]
    assert "color" in data["Collection"]


def test_calendar_service_get_calendars() -> None:
    """Test CalendarService get_calendars method."""
    mock_session = MagicMock(spec=PyiCloudSession)
    mock_response = MagicMock(spec=Response)
    mock_response.json.return_value = {"Collection": [{"title": "Test Calendar"}]}
    mock_session.get.return_value = mock_response
    with patch("pyicloud.services.calendar.get_localzone_name", return_value="UTC"):
        service = CalendarService(
            "https://example.com", mock_session, {"dsid": "12345"}
        )
        calendars = service.get_calendars()
        assert len(calendars) == 1
        assert calendars[0]["title"] == "Test Calendar"


def test_calendar_service_add_calendar() -> None:
    """Test CalendarService add_calendar method."""
    mock_session = MagicMock(spec=PyiCloudSession)
    mock_response = MagicMock(spec=Response)
    mock_response.json.return_value = {"status": "success"}
    mock_session.post.return_value = mock_response
    with patch("pyicloud.services.calendar.get_localzone_name", return_value="UTC"):
        service = CalendarService(
            "https://example.com", mock_session, {"dsid": "12345"}
        )
        calendar = CalendarObject(title="New Calendar")
        response = service.add_calendar(calendar)
        assert response["status"] == "success"


def test_calendar_service_remove_calendar() -> None:
    """Test CalendarService remove_calendar method."""
    mock_session = MagicMock(spec=PyiCloudSession)
    mock_response = MagicMock(spec=Response)
    mock_response.json.return_value = {"status": "success"}
    mock_session.post.return_value = mock_response

    with patch("pyicloud.services.calendar.get_localzone_name", return_value="UTC"):
        service = CalendarService(
            "https://example.com", mock_session, {"dsid": "12345"}
        )
        response = service.remove_calendar("calendar123")
        assert response["status"] == "success"


def test_calendar_service_get_events() -> None:
    """Test CalendarService get_events method."""
    mock_session = MagicMock(spec=PyiCloudSession)
    mock_response = MagicMock(spec=Response)
    mock_response.json.return_value = {"Event": [{"title": "Test Event"}]}
    mock_session.get.return_value = mock_response
    with patch("pyicloud.services.calendar.get_localzone_name", return_value="UTC"):
        service = CalendarService(
            "https://example.com", mock_session, {"dsid": "12345"}
        )
        events = service.get_events()
        assert len(events) == 1
        assert events[0]["title"] == "Test Event"


def test_calendar_service_add_event() -> None:
    """Test CalendarService add_event method."""
    mock_session = MagicMock(spec=PyiCloudSession)
    mock_response = MagicMock(spec=Response)
    mock_response.json.return_value = {"status": "success"}
    mock_session.post.return_value = mock_response
    with patch("pyicloud.services.calendar.get_localzone_name", return_value="UTC"):
        service = CalendarService(
            "https://example.com", mock_session, {"dsid": "12345"}
        )
        service.get_ctag = MagicMock(return_value="etag123")
        event = EventObject(pguid="calendar123", title="New Event")
        response = service.add_event(event)
        assert response["status"] == "success"


def test_calendar_service_remove_event() -> None:
    """Test CalendarService remove_event method."""
    mock_session = MagicMock(spec=PyiCloudSession)
    mock_response = MagicMock(spec=Response)
    mock_response.json.return_value = {"status": "success"}
    mock_session.post.return_value = mock_response
    with patch("pyicloud.services.calendar.get_localzone_name", return_value="UTC"):
        service = CalendarService(
            "https://example.com", mock_session, {"dsid": "12345"}
        )
        service.get_ctag = MagicMock(return_value="etag123")

        event = EventObject(pguid="calendar123", title="New Event")
        response = service.remove_event(event)
        assert response["status"] == "success"



"""Library base file."""

import base64
import getpass
import hashlib
import logging
from os import environ, mkdir, path
from tempfile import gettempdir
from typing import Any, Dict, List, Optional
from uuid import uuid1

import srp
from fido2.client import Fido2Client
from fido2.hid import CtapHidDevice
from fido2.webauthn import (
    AuthenticatorAssertionResponse,
    PublicKeyCredentialDescriptor,
    PublicKeyCredentialRequestOptions,
    PublicKeyCredentialType,
)
from requests import HTTPError
from requests.models import Response

from pyicloud.const import ACCOUNT_NAME, CONTENT_TYPE_JSON
from pyicloud.exceptions import (
    PyiCloud2FARequiredException,
    PyiCloudAPIResponseException,
    PyiCloudFailedLoginException,
    PyiCloudPasswordException,
    PyiCloudServiceNotActivatedException,
    PyiCloudServiceUnavailable,
)
from pyicloud.services import (
    AccountService,
    AppleDevice,
    CalendarService,
    ContactsService,
    DriveService,
    FindMyiPhoneServiceManager,
    HideMyEmailService,
    PhotosService,
    RemindersService,
    UbiquityService,
)
from pyicloud.session import PyiCloudSession
from pyicloud.utils import get_password_from_keyring

LOGGER: logging.Logger = logging.getLogger(__name__)


def b64url_decode(s: str) -> bytes:
    """Decode a base64url encoded string."""
    return base64.urlsafe_b64decode(s + "=" * (-len(s) % 4))


def b64_encode(b: bytes) -> str:
    """Encode bytes to a base64url encoded string."""
    return base64.b64encode(b).decode()


class SrpPassword:
    """SRP password."""

    def __init__(self, password: str) -> None:
        self.password: str = password
        self.salt: bytes
        self.iterations: int
        self.key_length: int

    def set_encrypt_info(self, salt: bytes, iterations: int, key_length: int) -> None:
        """Set encrypt info."""
        self.salt: bytes = salt
        self.iterations: int = iterations
        self.key_length: int = key_length

    def encode(self) -> bytes:
        """Encode password."""
        password_hash: bytes = hashlib.sha256(self.password.encode("utf-8")).digest()
        return hashlib.pbkdf2_hmac(
            "sha256",
            password_hash,
            self.salt,
            self.iterations,
            self.key_length,
        )


class PyiCloudPasswordFilter(logging.Filter):
    """Password log hider."""

    def __init__(self, password: str) -> None:
        super().__init__(password)

    def filter(self, record) -> bool:
        message: str = record.getMessage()
        if self.name in message:
            record.msg = message.replace(self.name, "*" * 8)
            record.args = ()

        return True


class PyiCloudService(object):
    """
    A base authentication class for the iCloud service. Handles the
    authentication required to access iCloud services.

    Usage:
        from pyicloud import PyiCloudService
        pyicloud = PyiCloudService('username@apple.com', 'password')
        pyicloud.iphone.location()
    """

    def _setup_endpoints(self) -> None:
        """Set up the endpoints for the service."""
        # If the country or region setting of your Apple ID is China mainland.
        # See https://support.apple.com/en-us/HT208351
        icloud_china: str = ".cn" if self._is_china_mainland else ""
        self.auth_endpoint: str = (
            f"https://idmsa.apple.com{icloud_china}/appleauth/auth"
        )
        self.home_endpoint: str = f"https://www.icloud.com{icloud_china}"
        self.setup_endpoint: str = f"https://setup.icloud.com{icloud_china}/setup/ws/1"

    def _setup_cookie_directory(self, cookie_directory: Optional[str]) -> str:
        """Set up the cookie directory for the service."""
        _cookie_directory: str = ""
        if cookie_directory:
            _cookie_directory = path.expanduser(path.normpath(cookie_directory))
            if not path.exists(_cookie_directory):
                mkdir(_cookie_directory, 0o700)
        else:
            topdir: str = path.join(gettempdir(), "pyicloud")
            _cookie_directory = path.join(topdir, getpass.getuser())
            if not path.exists(topdir):
                mkdir(topdir, 0o777)
            if not path.exists(_cookie_directory):
                mkdir(_cookie_directory, 0o700)

        return _cookie_directory

    def __init__(
        self,
        apple_id: str,
        password: Optional[str] = None,
        cookie_directory: Optional[str] = None,
        verify: bool = True,
        client_id: Optional[str] = None,
        with_family: bool = True,
        china_mainland: bool = False,
    ) -> None:
        self._is_china_mainland: bool = (
            china_mainland or environ.get("icloud_china", "0") == "1"
        )
        self._setup_endpoints()
        self._password: Optional[str] = password
        self._apple_id: str = apple_id

        if self._password is None:
            self._password = get_password_from_keyring(apple_id)

        self.data: dict[str, Any] = {}

        self.params: dict[str, Any] = {}
        self._client_id: str = client_id or (f"auth-{str(uuid1()).lower()}")
        self._with_family: bool = with_family
        self._password_filter: Optional[PyiCloudPasswordFilter] = None

        _cookie_directory = self._setup_cookie_directory(cookie_directory)

        self.session: PyiCloudSession = PyiCloudSession(
            self,
            verify=verify,
            headers={
                "Origin": self.home_endpoint,
                "Referer": f"{self.home_endpoint}/",
                "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/18.3.1 Safari/605.1.15",
            },
            client_id=self._client_id,
            cookie_directory=_cookie_directory,
        )

        self._client_id = self.session.data.get("client_id", self._client_id)

        self.params = {
            "clientBuildNumber": "2512Hotfix21",
            "clientMasteringNumber": "2512Hotfix21",
            "ckjsBuildVersion": "2310ProjectDev27",
            "ckjsVersion": "2.6.4",
            "clientId": self._client_id,
        }

        self._webservices: Optional[dict[str, dict[str, Any]]] = None

        self._account: Optional[AccountService] = None
        self._calendar: Optional[CalendarService] = None
        self._contacts: Optional[ContactsService] = None
        self._devices: Optional[FindMyiPhoneServiceManager] = None
        self._drive: Optional[DriveService] = None
        self._files: Optional[UbiquityService] = None
        self._hidemyemail: Optional[HideMyEmailService] = None
        self._photos: Optional[PhotosService] = None
        self._reminders: Optional[RemindersService] = None

        self.authenticate()

    def authenticate(
        self, force_refresh: bool = False, service: Optional[str] = None
    ) -> None:
        """
        Handles authentication, and persists cookies so that
        subsequent logins will not cause additional e-mails from Apple.
        """

        login_successful = False
        if self.session.data.get("session_token") and not force_refresh:
            try:
                self.data = self._validate_token()
                login_successful = True
            except PyiCloudAPIResponseException:
                LOGGER.debug("Invalid authentication token, will log in from scratch.")

        if (
            not login_successful
            and service is not None
            and service in self.data["apps"]
        ):
            app: dict[str, Any] = self.data["apps"][service]
            if "canLaunchWithOneFactor" in app and app["canLaunchWithOneFactor"]:
                LOGGER.debug("Authenticating as %s for %s", self.account_name, service)
                try:
                    self._authenticate_with_credentials_service(service)
                    login_successful = True
                except Exception:
                    LOGGER.debug(
                        "Could not log into service. Attempting brand new login."
                    )

        if not login_successful:
            try:
                self._authenticate()
            except PyiCloud2FARequiredException:
                LOGGER.debug("2FA is required")

        self._update_state()

    def _update_state(self) -> None:
        """Update the state of the service."""
        if (
            "dsInfo" in self.data
            and isinstance(self.data["dsInfo"], dict)
            and "dsid" in self.data["dsInfo"]
        ):
            self.params.update({"dsid": self.data["dsInfo"]["dsid"]})

        if "webservices" in self.data:
            self._webservices = self.data["webservices"]

            LOGGER.debug("Authentication completed successfully")

    def _authenticate(self) -> None:
        LOGGER.debug("Authenticating as %s", self.account_name)

        try:
            self._authenticate_with_token()
        except (PyiCloudFailedLoginException, PyiCloud2FARequiredException):
            headers: dict[str, str] = self._get_auth_headers()

            self._srp_authentication(headers)
            self._authenticate_with_token()

    def _srp_authentication(self, headers: dict[str, Any]) -> None:
        """SRP authentication."""
        try:
            srp_password = SrpPassword(self.password)
        except PyiCloudPasswordException as error:
            raise PyiCloudFailedLoginException("Password not provided") from error
        srp.rfc5054_enable()
        srp.no_username_in_x()
        usr = srp.User(
            self.account_name,
            srp_password,
            hash_alg=srp.SHA256,
            ng_type=srp.NG_2048,
        )
        uname, A = usr.start_authentication()  # pylint: disable=invalid-name
        data: dict[str, Any] = {
            "a": b64_encode(A),
            ACCOUNT_NAME: uname,
            "protocols": ["s2k", "s2k_fo"],
        }

        try:
            response: Response = self.session.post(
                f"{self.auth_endpoint}/signin/init",
                json=data,
                headers=headers,
            )
            response.raise_for_status()
        except (PyiCloudAPIResponseException, HTTPError) as error:
            msg = "Failed to initiate srp authentication."
            raise PyiCloudFailedLoginException(msg, error) from error

        body = response.json()
        salt: bytes = base64.b64decode(body["salt"])
        b: bytes = base64.b64decode(body["b"])
        c = body["c"]
        iterations: int = body["iteration"]
        key_length: int = 32
        srp_password.set_encrypt_info(salt, iterations, key_length)
        m1 = usr.process_challenge(salt, b)
        m2 = usr.H_AMK
        if m1 and m2:
            data = {
                ACCOUNT_NAME: uname,
                "c": c,
                "m1": b64_encode(m1),
                "m2": b64_encode(m2),
                "rememberMe": True,
                "trustTokens": [],
            }
        if self.session.data.get("trust_token"):
            data["trustTokens"] = [self.session.data.get("trust_token")]

        try:
            self.session.post(
                f"{self.auth_endpoint}/signin/complete",
                params={
                    "isRememberMeEnabled": "true",
                },
                json=data,
                headers=headers,
            )
        except PyiCloud2FARequiredException:
            LOGGER.debug("2FA required to complete authentication.")
        except PyiCloudAPIResponseException as error:
            msg = "Invalid email/password combination."
            raise PyiCloudFailedLoginException(msg) from error

    def _authenticate_with_token(self) -> None:
        """Authenticate using session token."""
        if not self.session.data.get("session_token"):
            raise PyiCloudFailedLoginException("No session token available")

        try:
            data: dict[str, Any] = {
                "accountCountryCode": self.session.data.get("account_country"),
                "dsWebAuthToken": self.session.data.get("session_token"),
                "extended_login": True,
                "trustToken": self.session.data.get("trust_token", ""),
            }

            resp: Response = self.session.post(
                f"{self.setup_endpoint}/accountLogin", json=data
            )
            resp.raise_for_status()

            self.data = resp.json()
            if not self.is_trusted_session:
                raise PyiCloud2FARequiredException(self.account_name, resp)
        except (PyiCloudAPIResponseException, HTTPError) as error:
            msg = "Invalid authentication token."
            raise PyiCloudFailedLoginException(msg, error) from error

    def _authenticate_with_credentials_service(self, service: Optional[str]) -> None:
        """Authenticate to a specific service using credentials."""
        data: dict[str, Any] = {
            "appName": service,
            "apple_id": self.account_name,
            "password": self.password,
        }

        try:
            self.session.post(f"{self.setup_endpoint}/accountLogin", json=data)

            self.data = self._validate_token()
        except PyiCloudAPIResponseException as error:
            msg = "Invalid email/password combination."
            raise PyiCloudFailedLoginException(msg, error) from error

    def _validate_token(self) -> Any:
        """Checks if the current access token is still valid."""
        LOGGER.debug("Checking session token validity")
        if not self.session.cookies.get("X-APPLE-WEBAUTH-TOKEN"):
            raise PyiCloudAPIResponseException(
                "Missing X-APPLE-WEBAUTH-TOKEN cookie", None
            )
        try:
            req: Response = self.session.post(
                f"{self.setup_endpoint}/validate", data="null"
            )
            LOGGER.debug("Session token is still valid")
            return req.json()
        except PyiCloudAPIResponseException:
            LOGGER.debug("Invalid authentication token")
            raise

    def _get_auth_headers(
        self, overrides: Optional[dict[str, Any]] = None
    ) -> dict[str, Any]:
        headers: dict[str, Any] = {
            "Accept": f"{CONTENT_TYPE_JSON}, text/javascript",
            "Content-Type": CONTENT_TYPE_JSON,
            "X-Apple-OAuth-Client-Id": "d39ba9916b7251055b22c7f910e2ea796ee65e98b2ddecea8f5dde8d9d1a815d",
            "X-Apple-OAuth-Client-Type": "firstPartyAuth",
            "X-Apple-OAuth-Redirect-URI": "https://www.icloud.com",
            "X-Apple-OAuth-Require-Grant-Code": "true",
            "X-Apple-OAuth-Response-Mode": "web_message",
            "X-Apple-OAuth-Response-Type": "code",
            "X-Apple-OAuth-State": self._client_id,
            "X-Apple-Widget-Key": "d39ba9916b7251055b22c7f910e2ea796ee65e98b2ddecea8f5dde8d9d1a815d",
        }

        if self.session.data.get("scnt"):
            headers["scnt"] = self.session.data["scnt"]

        if self.session.data.get("session_id"):
            headers["X-Apple-ID-Session-Id"] = self.session.data["session_id"]

        if overrides:
            headers.update(overrides)

        return headers

    @property
    def requires_2sa(self) -> bool:
        """Returns True if two-step authentication is required."""
        return self.data.get("dsInfo", {}).get("hsaVersion", 0) >= 1 and (
            self.data.get("hsaChallengeRequired", False) or not self.is_trusted_session
        )

    @property
    def requires_2fa(self) -> bool:
        """Returns True if two-factor authentication is required."""
        return self.data.get("dsInfo", {}).get("hsaVersion", 0) == 2 and (
            self.data.get("hsaChallengeRequired", False) or not self.is_trusted_session
        )

    @property
    def is_trusted_session(self) -> bool:
        """Returns True if the session is trusted."""
        return self.data.get("hsaTrustedBrowser", False)

    @property
    def trusted_devices(self) -> list[dict[str, Any]]:
        """Returns devices trusted for two-step authentication."""
        request: Response = self.session.get(
            f"{self.setup_endpoint}/listDevices", params=self.params
        )
        return request.json().get("devices")

    def send_verification_code(self, device: dict[str, Any]) -> bool:
        """Requests that a verification code is sent to the given device."""
        request: Response = self.session.post(
            f"{self.setup_endpoint}/sendVerificationCode",
            params=self.params,
            json=device,
        )
        return request.json().get("success", False)

    def validate_verification_code(self, device: dict[str, Any], code: str) -> bool:
        """Verifies a verification code received on a trusted device."""
        device.update({"verificationCode": code, "trustBrowser": True})

        try:
            self.session.post(
                f"{self.setup_endpoint}/validateVerificationCode",
                params=self.params,
                json=device,
            )
        except PyiCloudAPIResponseException as error:
            if error.code == -21669:
                # Wrong verification code
                return False
            raise

        self.trust_session()

        return not self.requires_2sa

    def _get_webauthn_options(self) -> Dict:
        """Retrieve WebAuthn request options (PublicKeyCredentialRequestOptions) for assertion."""
        headers = self._get_auth_headers({"Accept": CONTENT_TYPE_JSON})

        return self.session.get(self.auth_endpoint, headers=headers).json()

    @property
    def security_key_names(self) -> Optional[List[str]]:
        """Security key names which can be used for the WebAuthn assertion."""
        return self._get_webauthn_options().get("keyNames")

    def _submit_webauthn_assertion_response(self, data: Dict) -> None:
        """Submit the WebAuthn assertion response for authentication."""
        headers = self._get_auth_headers({"Accept": CONTENT_TYPE_JSON})

        self.session.post(
            f"{self.auth_endpoint}/verify/security/key", json=data, headers=headers
        )

    @property
    def fido2_devices(self) -> List[CtapHidDevice]:
        """List the available FIDO2 devices."""
        return list(CtapHidDevice.list_devices())

    def confirm_security_key(self, device: Optional[CtapHidDevice] = None) -> None:
        """Conduct the WebAuthn assertion ceremony with user's FIDO2 device."""
        options = self._get_webauthn_options()
        challenge = options["fsaChallenge"]["challenge"]
        allowed_credentials = options["fsaChallenge"]["keyHandles"]
        rp_id = options["fsaChallenge"]["rpId"]

        if not device:
            devices: List[CtapHidDevice] = list(CtapHidDevice.list_devices())

            if not devices:
                raise RuntimeError("No FIDO2 devices found")

            device = devices[0]

        client = Fido2Client(
            device=device,
            origin="https://apple.com",
        )
        credentials: List[PublicKeyCredentialDescriptor] = [
            PublicKeyCredentialDescriptor(
                id=b64url_decode(cred_id), type=PublicKeyCredentialType("public-key")
            )
            for cred_id in allowed_credentials
        ]
        assertion_options = PublicKeyCredentialRequestOptions(
            challenge=b64url_decode(challenge),
            rp_id=rp_id,
            allow_credentials=credentials,
        )
        response: AuthenticatorAssertionResponse = client.get_assertion(
            assertion_options
        ).get_response(0)

        self._submit_webauthn_assertion_response(
            {
                "challenge": challenge,
                "clientData": b64_encode(response.client_data),
                "signatureData": b64_encode(response.signature),
                "authenticatorData": b64_encode(response.authenticator_data),
                "userHandle": b64_encode(response.user_handle)
                if response.user_handle
                else None,
                "credentialID": b64_encode(response.credential_id)
                if response.credential_id
                else None,
                "rpId": rp_id,
            }
        )

        self.trust_session()

    def validate_2fa_code(self, code: str) -> bool:
        """Verifies a verification code received via Apple's 2FA system (HSA2)."""
        data: dict[str, Any] = {"securityCode": {"code": code}}
        headers: dict[str, Any] = self._get_auth_headers({"Accept": CONTENT_TYPE_JSON})

        try:
            self.session.post(
                f"{self.auth_endpoint}/verify/trusteddevice/securitycode",
                json=data,
                headers=headers,
            )
        except PyiCloudAPIResponseException:
            # Wrong verification code
            LOGGER.error("Code verification failed.")
            return False

        LOGGER.debug("Code verification successful.")

        self.trust_session()
        return not self.requires_2sa

    def trust_session(self) -> bool:
        """Request session trust to avoid user log in going forward."""
        headers: dict[str, Any] = self._get_auth_headers()

        try:
            self.session.get(
                f"{self.auth_endpoint}/2sv/trust",
                headers=headers,
            )
            self._authenticate_with_token()
            return True
        except (PyiCloudAPIResponseException, PyiCloud2FARequiredException):
            LOGGER.error("Session trust failed.")
            return False

    def get_webservice_url(self, ws_key: str) -> str:
        """Get webservice URL, raise an exception if not exists."""
        if self._webservices is None or self._webservices.get(ws_key) is None:
            raise PyiCloudServiceNotActivatedException(
                f"Webservice not available: {ws_key}"
            )

        return self._webservices[ws_key]["url"]

    @property
    def devices(self) -> FindMyiPhoneServiceManager:
        """Returns all devices."""
        if not self._devices:
            try:
                service_root: str = self.get_webservice_url("findme")
                self._devices = FindMyiPhoneServiceManager(
                    service_root, self.session, self.params, self._with_family
                )
            except PyiCloudServiceNotActivatedException as error:
                raise PyiCloudServiceUnavailable(
                    "Find My iPhone service not available"
                ) from error
        return self._devices

    @property
    def hidemyemail(self) -> HideMyEmailService:
        """Gets the 'HME' service."""
        if not self._hidemyemail:
            service_root: str = self.get_webservice_url("premiummailsettings")
            try:
                self._hidemyemail = HideMyEmailService(
                    service_root, self.session, self.params
                )
            except PyiCloudAPIResponseException as error:
                raise PyiCloudServiceUnavailable(
                    "Hide My Email service not available"
                ) from error
        return self._hidemyemail

    @property
    def iphone(self) -> AppleDevice:
        """Returns the iPhone."""
        return self.devices[0]

    @property
    def account(self) -> AccountService:
        """Gets the 'Account' service."""
        if not self._account:
            service_root: str = self.get_webservice_url("account")
            try:
                self._account = AccountService(
                    service_root=service_root,
                    session=self.session,
                    china_mainland=self._is_china_mainland,
                    params=self.params,
                )
            except (PyiCloudAPIResponseException,) as error:
                raise PyiCloudServiceUnavailable(
                    "Account service not available"
                ) from error
        return self._account

    @property
    def files(self) -> UbiquityService:
        """Gets the 'File' service."""
        if not self._files:
            service_root: str = self.get_webservice_url("ubiquity")
            try:
                self._files = UbiquityService(service_root, self.session, self.params)
            except (PyiCloudAPIResponseException,) as error:
                raise PyiCloudServiceUnavailable(
                    "Files service not available"
                ) from error
        return self._files

    @property
    def photos(self) -> PhotosService:
        """Gets the 'Photo' service."""
        if not self._photos:
            service_root: str = self.get_webservice_url("ckdatabasews")
            upload_url: str = self.get_webservice_url("uploadimagews")
            shared_streams_url: str = self.get_webservice_url("sharedstreams")
            self.params["dsid"] = self.data["dsInfo"]["dsid"]

            try:
                self._photos = PhotosService(
                    service_root,
                    self.session,
                    self.params,
                    upload_url,
                    shared_streams_url,
                )
            except (PyiCloudAPIResponseException,) as error:
                raise PyiCloudServiceUnavailable(
                    "Photos service not available"
                ) from error
        return self._photos

    @property
    def calendar(self) -> CalendarService:
        """Gets the 'Calendar' service."""
        if not self._calendar:
            service_root: str = self.get_webservice_url("calendar")
            try:
                self._calendar = CalendarService(
                    service_root, self.session, self.params
                )
            except (PyiCloudAPIResponseException,) as error:
                raise PyiCloudServiceUnavailable(
                    "Calendar service not available"
                ) from error
        return self._calendar

    @property
    def contacts(self) -> ContactsService:
        """Gets the 'Contacts' service."""
        if not self._contacts:
            service_root: str = self.get_webservice_url("contacts")
            try:
                self._contacts = ContactsService(
                    service_root, self.session, self.params
                )
            except (PyiCloudAPIResponseException,) as error:
                raise PyiCloudServiceUnavailable(
                    "Contacts service not available"
                ) from error
        return self._contacts

    @property
    def reminders(self) -> RemindersService:
        """Gets the 'Reminders' service."""
        if not self._reminders:
            service_root: str = self.get_webservice_url("reminders")
            try:
                self._reminders = RemindersService(
                    service_root, self.session, self.params
                )
            except (PyiCloudAPIResponseException,) as error:
                raise PyiCloudServiceUnavailable(
                    "Reminders service not available"
                ) from error
        return self._reminders

    @property
    def drive(self) -> DriveService:
        """Gets the 'Drive' service."""
        if not self._drive:
            try:
                self._drive = DriveService(
                    service_root=self.get_webservice_url("drivews"),
                    document_root=self.get_webservice_url("docws"),
                    session=self.session,
                    params=self.params,
                )
            except (PyiCloudAPIResponseException,) as error:
                raise PyiCloudServiceUnavailable(
                    "Drive service not available"
                ) from error
        return self._drive

    @property
    def account_name(self) -> str:
        """Retrieves the account name associated with the Apple ID."""

        return self._apple_id

    @property
    def password(self) -> str:
        """Retrieves the password associated with the Apple ID."""
        if self._password is None:
            raise PyiCloudPasswordException()
        self._password_filter = PyiCloudPasswordFilter(self._password)
        LOGGER.addFilter(self._password_filter)
        return self._password

    @property
    def password_filter(self) -> Optional[PyiCloudPasswordFilter]:
        """Retrieves the password filter"""
        return self._password_filter if self._password_filter else None

    def __str__(self) -> str:
        return f"iCloud API: {self.account_name}"

    def __repr__(self) -> str:
        return f"<{self}>"
